from abc import ABC, abstractclassmethod
from inspect import iscoroutinefunction
from typing import Awaitable, Callable

from promplate.llm.base import *
from promplate.prompt import ChatTemplate, Template
from promplate.prompt.template import Context

from .utils import appender, count_position_parameters

PreProcess = Callable[[Context], Context]
PostProcess = PreProcess | Callable[[Context, str], Context]

AsyncPreProcess = Callable[[Context], Awaitable[Context]]
AsyncPostProcess = AsyncPreProcess | Callable[[Context, str], Awaitable[Context]]


class AbstractChain(ABC):
    @abstractclassmethod
    def run(self, context: Context, complete: Complete | None):
        pass

    @abstractclassmethod
    async def arun(self, context: Context, complete: Complete | AsyncComplete | None):
        pass


class Node(AbstractChain):
    def __init__(
        self,
        template: Template,
        pre_processes: list[PreProcess | AsyncPreProcess] | None = None,
        post_processes: list[PostProcess | AsyncPostProcess] | None = None,
        complete: Complete | AsyncComplete | None = None,
        **config,
    ):
        self.template = template
        self.pre_processes = pre_processes or []
        self.post_processes = post_processes or []
        self.complete = complete
        self.run_config = config

    @property
    def pre_process(self):
        return appender(self.pre_processes)

    @property
    def post_process(self):
        return appender(self.post_processes)

    @staticmethod
    def via(
        process: PostProcess | AsyncPostProcess,
        context: Context,
        result: str,
    ) -> Context | Awaitable[Context]:
        if count_position_parameters(process) == 1:
            context = process(context) or context
        else:
            context = process(context, result) or context

    def run(self, context, complete=None):
        complete = complete or self.complete
        assert complete is not None

        for process in self.pre_processes:
            context = process(context) or context

        prompt = self.template.render(context)

        assert isinstance(self.template, ChatTemplate) ^ isinstance(prompt, str)

        result = context["__result__"] = complete(prompt, **self.run_config)

        for process in self.post_processes:
            context = self.via(process, context, result) or context

        return context

    async def arun(self, context, complete=None):
        complete = complete or self.complete
        assert complete is not None

        for process in self.pre_processes:
            if iscoroutinefunction(process):
                context = await process(context) or context
            else:
                context = process(context) or context

        prompt = await self.template.arender(context)

        assert isinstance(self.template, ChatTemplate) ^ isinstance(prompt, str)

        if iscoroutinefunction(complete):
            result = context["__result__"] = await complete(prompt, **self.run_config)
        else:
            result = context["__result__"] = complete(prompt, **self.run_config)

        for process in self.post_processes:
            if iscoroutinefunction(process):
                context = await self.via(process, context, result) or context
            else:
                context = self.via(process, context, result) or context

        return context

    def next(self, chain: AbstractChain):
        if isinstance(chain, Node):
            return Chain(self, chain)
        else:
            return Chain(self, *chain)

    def __add__(self, chain: AbstractChain):
        return self.next(chain)


class Chain(AbstractChain):
    def __init__(
        self,
        *nodes: AbstractChain,
        complete: Complete | AsyncComplete | None = None,
    ):
        self.nodes = list(nodes)
        self.complete = complete

    def next(self, chain: AbstractChain):
        if isinstance(chain, Node):
            return Chain(*self, chain)
        else:
            return Chain(*self, *chain)

    def __add__(self, chain):
        return self.next(chain)

    def __iter__(self):
        return iter(self.nodes)

    def _run(self, context, complete):
        for node in self.nodes:
            context = node.run(context, complete)

        return context

    def run(self, context, complete=None):
        complete = complete or self.complete
        try:
            return self._run(context, complete)
        except JumpTo as jump:
            return jump.chain.run(jump.context or context, complete)

    async def _arun(self, context, complete):
        for node in self.nodes:
            context = await node.arun(context, complete)

        return context

    async def arun(self, context, complete=None):
        complete = complete or self.complete
        try:
            return await self._arun(context, complete)
        except JumpTo as jump:
            return await jump.chain.arun(jump.context or context, complete)


class JumpTo(Exception):
    def __init__(self, chain: Node | Chain, context: Context | None = None):
        self.chain = chain
        self.context = context



from openai import ChatCompletion, Completion
from promplate.prompt.chat import Message

from .base import *


def text_complete(**default_config) -> CompleteText:
    def complete_text(text: str, **config):
        config = default_config | config | {"stream": False, "prompt": text}
        result = Completion.create(**config)
        return result["choices"][0]["text"]

    return complete_text


def async_text_complete(**default_config) -> AsyncCompleteText:
    async def async_complete_text(text: str, **config):
        config = default_config | config | {"stream": False, "prompt": text}
        result = await Completion.acreate(**config)
        return result["choices"][0]["text"]

    return async_complete_text


def text_generate(**default_config) -> GenerateText:
    def generate_text(text: str, **config):
        config = default_config | config | {"stream": True, "prompt": text}
        stream = Completion.create(**config)
        for event in stream:
            yield event["choices"][0]["text"]

    return generate_text


def async_text_generate(**default_config) -> AsyncGenerateText:
    async def async_generate_text(text: str, **config):
        config = default_config | config | {"stream": True, "prompt": text}
        stream = Completion.acreate(**config)
        async for event in stream:
            yield event["choices"][0]["text"]

    return async_generate_text


def chat_complete(**default_config) -> CompleteChat:
    def complete_chat(messages: list[Message], **config):
        config = default_config | config | {"stream": False, "messages": messages}
        result = ChatCompletion.create(**config)
        return result["choices"][0]["message"]["content"]

    return complete_chat


def async_chat_complete(**default_config) -> AsyncCompleteChat:
    async def async_complete_chat(messages: list[Message], **config):
        config = default_config | config | {"stream": False, "messages": messages}
        result = await ChatCompletion.acreate(**config)
        return result["choices"][0]["message"]["content"]

    return async_complete_chat


def chat_generate(**default_config) -> GenerateChat:
    def generate_chat(messages: list[Message], **config):
        config = default_config | config | {"stream": True, "messages": messages}
        stream = ChatCompletion.create(**config)
        for event in stream:
            delta: dict = event["choices"][0]["delta"]
            yield delta.get("content", "")

    return generate_chat


def async_chat_generate(**default_config) -> AsyncGenerateChat:
    async def async_generate_chat(messages: list[Message], **config):
        config = default_config | config | {"stream": True, "messages": messages}
        stream = ChatCompletion.acreate(**config)
        async for event in stream:
            delta: dict = event["choices"][0]["delta"]
            yield delta.get("content", "")

    return async_generate_chat



from typing import AsyncIterator, Awaitable, Callable, Iterator

from promplate.prompt.chat import Message

CompleteText = Callable[[str], str]
CompleteChat = Callable[[list[Message]], str]
GenerateText = Callable[[str], Iterator[str]]
GenerateChat = Callable[[list[Message]], Iterator[str]]
Complete = CompleteText | CompleteChat
Generate = GenerateText | GenerateChat

AsyncCompleteText = Callable[[str], Awaitable[str]]
AsyncCompleteChat = Callable[[list[Message]], Awaitable[str]]
AsyncGenerateText = Callable[[str], AsyncIterator[str]]
AsyncGenerateChat = Callable[[list[Message]], AsyncIterator[str]]
AsyncComplete = AsyncCompleteText | AsyncCompleteChat
AsyncGenerate = AsyncGenerateText | AsyncGenerateChat
